print "Encrypt and decrypt ADFGVX v 1.0 Feb 11, 2017"
print "main functions:"
print "encrypt_ADFGVX(plaintext, key1, key2) - encrypt with ADFGVX"
print "decrypt_ADFGVX(cypertext, key1, key2) - decrypt with ADFGVX"
print "random_ADFGVX_key1() - generate a random ADFGVX (6x6 square) as a string"
print "random_ADFGVX_key2(n) - generate a random permutation of n (n should be even)\n"

alpha = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789" # the English alphabet
ADFGVX = "ADFGVX"

def clean_text_ADFGVX(ptext, spaces=False):
    """
    Change lowercase letters to upper case and delete spaces and punctuation
    put text into 5 grams

    INPUT:

    - ``ptext`` -- a string

    OUTPUT:

    - a string with only uppercase letters and spaces

    EXAMPLES::

        sage: clean_text_ADFGVX("Now * the TIME :9: for all good !")
        'NOWTHETIME9FORALLGOOD'
        sage: clean_text_ADFGVX("Now * the TIME :: for all good !",True)
        'NOWTH ETIME 9FORA LLGOO D'
    """
    out = ""
    cnt = 0
    for i in range(len(ptext)):
        if ptext[i].isalnum():
            if not mod(cnt,5) and cnt>0 and spaces:
                out+=" "
            out += ptext[i].upper()
            cnt += 1
    return out

def letter_to_ADFGVX_pair(let, key1):
    """
    Change a letter to a pair of 'ADFGVX' letters using ``key1``
    """
    n = key1.index(let.upper())
    return ADFGVX[n//6]+ADFGVX[mod(n,6)]

def ADFGVX_pair_to_letter(pair, key1):
    """
    Change a pair of 'ADFGVX' letters into a letter using ``key1``
    """
    n = 6*ADFGVX.index(pair[0])+ADFGVX.index(pair[1])
    return key1[n]

def apply_key1(plaintext, key1):
    """
    Replace each letter in the plaintext with a pair of letters from ADFGVX
    """
    out = ""
    for let in plaintext:
        if let.isalnum():
            out+= letter_to_ADFGVX_pair(let, key1)+" "
    return out

def unapply_key1(cyphertext, key1):
    """
    Replace each pair of letters of ADFGVX with a letter from the key
    """
    cctext = clean_text_ADFGVX(cyphertext)
    out = ""
    for i in range(len(cctext)//2):
        out += ADFGVX_pair_to_letter(cctext[2*i]+cctext[2*i+1], key1)
    return clean_text_ADFGVX(out, True)

def encrypt_rt(plaintext, perm):
    """
    Encrypt plaintext using rectangular transposition encypherment scheme.

    If the length of the key doesn't divide the length of the cyphertext
    we will pad the plaintext with repeated characters from the beginning
    so that it does divide.

    INPUT:

    - ``plaintext`` - a string
    - ``perm`` - a permutation or a list of distinct integers

    OUTPUT:

    - a string with only uppercase letters and spaces

    EXAMPLES::

        sage: encrypt_rt("Abcde fg",[1])
        'ABCDE FG'
        sage: encrypt_rt("Abcde fg",[1,2])
        'ACEGB DFA'
        sage: encrypt_rt("Abcde fg",[1,2,3,4,5,6,7,8])
        'ABCDE FGA'
        sage: encrypt_rt("Abcde fg",[8,7,6,5,4,3,2,1])
        'AGFED CBA'
        sage: encrypt_rt("Now is the time for all good",[2,1,3,4,5])
        'ONWIS HTETI EMFOR LALGO DONOW'
        
    """
    cl_ptext = clean_text_ADFGVX(plaintext,False)
    N = len(cl_ptext)
    cl_ptext += cl_ptext[:mod(-N,len(perm))]
    N = len(cl_ptext)
    out = ""
    iperm = Permutation(perm).inverse()
    for j in range(N/len(perm)):
        for i in range(len(perm)):
            if not mod(j*len(perm)+i,5) and j*len(perm)+i>0:
                out+=" "
            out+=cl_ptext[j*len(perm)+iperm[i]-1]
    return out

def decrypt_rt(cyphertext, perm):
    """
    Decrypt cyphertext using rectangular transposition encypherment scheme.

    INPUT:
    - ``cyphertext`` - a string
    - ``perm`` - a permutation or a list of distinct integers

    OUTPUT:

    - a string with only uppercase letters and spaces

    EXAMPLES::

        sage: decrypt_rt(encrypt_rt("Abcde fg",[8,7,6,5,4,3,2,1]),[8,7,6,5,4,3,2,1])
        'ABCDE FGA'
        sage: decrypt_rt(encrypt_rt("Now is the time for all good",[5,1,4,2,3]),[5,1,4,2,3])
        'NOWIS THETI MEFOR ALLGO ODNOW'
    """
    iperm = Permutation(perm).inverse()
    return encrypt_rt(cyphertext, iperm)

def random_ADFGVX_key1():
    """
    Generate a random key of length 36 which is a permutation of the alphabet/
    numbers 0-9.
    """
    p = Permutations(36).random_element()
    return "".join([alpha[i-1] for i in p])

def random_ADFGVX_key2(n):
    """
    Generate a random permutation of length n
    """
    return list(Permutations(n).random_element())

def encrypt_ADFGVX(plaintext, key1, key2):
    """
    encrypt using ADFGVX encypherment scheme
    """
    return encrypt_rt(apply_key1(plaintext, key1), key2)

def decrypt_ADFGVX(cyphertext, key1, key2):
    """
    decrypt using ADFGVX encypherment scheme
    """
    return unapply_key1(decrypt_rt(cyphertext, key2), key1)
